/* Copyright © 2012, 2013 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.http.headers;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

/**
 * Parse an accept field header string to a structured representation.
 * 
 * @author Matt Champion
 * @since 0.0.13
 */
public class ContentTypeParser implements HeaderParser<QContentType>
{
    public ContentTypeParser()
    {
    }

    @Override
    public boolean isCorrectHeader(String header)
    {
        return "ACCEPT".equals(header.toUpperCase());
    }

    @Override
    public Collection<QContentType> parseAll(String header)
            throws UnParsableHeaderException
    {
        if (header == null)
        {
            throw new UnParsableHeaderException("Null Pointer.");
        }
        Collection<QContentType> qCT = new ArrayList<QContentType>();
        if ("".equals(header))
        {
            return qCT;
        }
        String[] types = header.split(",");
        for (int i = 0; i < types.length; i++)
        {
            QContentType ct = parse(types[i].trim());
            if(ct != null)
            {
                qCT.add(ct);
            }
        }
        return qCT;
    }

    @Override
    public QContentType parse(String types) throws UnParsableHeaderException
    {
        if (types == null)
        {
            throw new UnParsableHeaderException("Null pointer");
        }
        if ("".equals(types.trim()))
        {
            throw new UnParsableHeaderException("Empty string");
        }
        String[] typeParts = types.trim().split("/");
        if (typeParts.length == 2)
        {
            String[] lastParts = typeParts[1].trim().split(";",-1);
            Map<String, String> map = new HashMap<String, String>();
            double qualifier = 1.0;
            if (lastParts.length > 1)
            {
                for (int i = 1; i < lastParts.length; i++)
                {
                    String[] extensions = lastParts[i].trim().split("=",-1);
                    if (extensions.length != 2)
                    {
                        throw new UnParsableHeaderException("Invalid parameter.");
                    }
                    if ("q".equals(extensions[0].trim()))
                    {
                        try
                        {
                            qualifier = Double.parseDouble(extensions[1].trim());
                        }
                        catch(NumberFormatException exception)
                        {
                            throw new UnParsableHeaderException("Unable to parse number");
                        }
                        if (qualifier < 0.0 || qualifier > 1.0)
                        {
                            throw new UnParsableHeaderException(
                                    "Invalid qualifier range.");
                        }
                    }
                    else
                    {
                        map.put(extensions[0].trim(), extensions[1].trim());
                    }
                }
            }
            else if (lastParts.length == 0)
            {
                throw new UnParsableHeaderException("Unparsable content type: "
                        + typeParts[1]);
            }
            return new QContentType(typeParts[0].trim(), lastParts[0].trim(), map, qualifier);
        }
        else
        {
            throw new UnParsableHeaderException("Unparsable content type: "
                    + types.trim() + ": " + typeParts.length);
        }
    }
}
