/* Copyright © 2012, 2013 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
 * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.http.headers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

/**
 * A parsed accept header field value with a qualification value.
 * <P>
 * The to string value is calculated when first accessed and cached subsequently. The value is checked and set
 * atomically. Concurrent access may cause different String objects to be cached and returned but they will be equal. A
 * happens before relationship is not established by calls to {@link toString()}.
 * 
 * @author Matt Champion
 * @since 0.0.13
 */
public class QContentType implements HeaderFieldElement, Qualified, Cloneable
{

    public static final QContentType TEXT_HTML = new QContentType("text", "html", null, 1.0);
    public static final QContentType APP_XHTML = new QContentType("application", "xhtml+xml", null,
            1.0);
    public static final QContentType APP_RDF = new QContentType("application", "rdf+xml", null, 1.0);

    private static final int HASH_SEED = 13;
    private static final int HASH_MULT = 5;
    private static final String WILDCARD = "*";

    private final String type;
    private final String subType;
    private final Map<String, String> parameters;
    private final double qualifier;
    
    private String toString;

    /**
     * Constructor for qualified content types.
     * <P>
     * Creates an immutable object. The type and subtype may not be null. The value of the qualifier
     * must be between 0.0 and 1.0.
     * 
     * @param type
     *            Type of the content type
     * @param subType
     *            Subtype of the content type
     * @param parameters
     *            Map of additional parameters
     * @param qualifier
     *            The qualifier of the content type
     * @throws IllegalArgumentException
     *             If type or subtype are null or if the qualifier is less than zero or greater than
     *             one
     */
    public QContentType(String type, String subType, Map<String, String> parameters,
            double qualifier) throws IllegalArgumentException
    {
        if (type == null)
        {
            throw new IllegalArgumentException("Type must not be null.");
        }
        else if (subType == null)
        {
            throw new IllegalArgumentException("SubType must not be null.");
        }
        else if (qualifier < 0.0 || qualifier > 1.0)
        {
            throw new IllegalArgumentException("Qualifier must be between 0.0 and 1.0");
        }
        else
        {
            this.type = type;
            this.subType = subType;
            if (parameters != null)
            {
                this.parameters = Collections.unmodifiableMap(new HashMap<String, String>(
                        parameters));
            }
            else
            {
                this.parameters = null;
            }
            this.qualifier = qualifier;
        }
    }

    /**
     * Parser for HttpServletRequest to QContentType
     * 
     * @param request
     *            The HTTP request to parse
     * @return The List of content types in the accept header.
     * @throws UnParsableHeaderException
     * @since 0.0.13
     */
    public static List<QContentType> getRequestContentTypes(HttpServletRequest request)
            throws UnParsableHeaderException
    {
        @SuppressWarnings("unchecked")
        Enumeration<String> headerNames = request.getHeaderNames();
        ContentTypeParser parser = new ContentTypeParser();
        List<QContentType> types = new ArrayList<QContentType>();
        while (headerNames.hasMoreElements())
        {
            String headerName = headerNames.nextElement();
            if (parser.isCorrectHeader(headerName))
            {
                String headers = request.getHeader(headerName);
                types.addAll(parser.parseAll(headers));
            }
        }
        return types;
    }

    @Override
    public String toString()
    {
        String toString = this.toString;
        if (toString != null)
        {
            return toString;
        }
        else
        {
            toString = type + "/" + subType + ";q=" + getQualifier();
            if (parameters != null)
            {
                for (String name : parameters.keySet())
                {
                    toString += ";" + name + "=" + parameters.get(name);
                }
            }
            this.toString = toString;
            return toString;
        }
    }

    @Override
    public double getQualifier()
    {
        return qualifier;
    }

    /**
     * Method that returns true if the two content types can be matched against one another. Takes
     * into account wildcard matching.
     * 
     * @param qct
     * @return True if they match
     * @since 0.0.13
     */
    public boolean sameType(QContentType qct)
    {
        if (type.equals(qct.type) && subType.equals(qct.subType))
        {
            return true;
        }
        else if (type.equals(qct.type) && (WILDCARD.equals(subType) || WILDCARD.equals(qct.subType)))
        {
            return true;
        }
        else if ((WILDCARD.equals(type) || WILDCARD.equals(qct.type)) && subType.equals(qct.subType))
        {
            return true;
        }
        else
        {
            return ((WILDCARD.equals(type) || WILDCARD.equals(qct.type)) && (WILDCARD.equals(subType)
                    || WILDCARD.equals(qct.subType)));
        }
    }

    /**
     * Test to see if the subtype is a wildcard.
     * 
     * @return True if the sub type is a wildcard
     * @since 0.0.13
     */
    public boolean isAnySubType()
    {
        return WILDCARD.equals(subType);
    }

    /**
     * Test to see if the type is a wildcard.
     * 
     * @return True if the type is a wild card.
     * @since 0.0.13
     */
    public boolean isAnyType()
    {
        return WILDCARD.equals(type);
    }

    @Override
    public int hashCode()
    {
        int hashCode = HASH_SEED;
        hashCode = (hashCode * HASH_MULT) + type.hashCode();
        hashCode = (hashCode * HASH_MULT) + subType.hashCode();
        if (parameters != null)
        {
            for (String key : parameters.keySet())
            {
                hashCode = (hashCode * HASH_MULT) + key.hashCode();
                hashCode = (hashCode * HASH_MULT) + parameters.get(key).hashCode();
            }
        }
        long temp = Double.doubleToLongBits(qualifier);
        hashCode = (hashCode * HASH_MULT) + (int) (temp ^ (temp >>> 32));
        return hashCode;
    }

    @Override
    public boolean equals(Object object)
    {
        if (object == null)
        {
            return false;
        }
        else if (this == object)
        {
            return true;
        }
        else if (object instanceof QContentType)
        {
            QContentType contentType = (QContentType) object;
            if (!(this.type.equals(contentType.type) && this.subType.equals(contentType.subType)
                    && this.qualifier == contentType.qualifier))
            {
                return false;
            }
            else
            {
                if (parameters == null)
                {
                    if (contentType.parameters == null)
                    {
                        return true;
                    }
                    else if (contentType.parameters.size() == 0)
                    {
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else
                {
                    if (contentType.parameters == null)
                    {
                        if (parameters.size() == 0)
                        {
                            return true;
                        }
                        else
                        {
                            return false;
                        }
                    }
                    else if (parameters.size() == contentType.parameters.size())
                    {
                        Set<String> keys = parameters.keySet();
                        for (String key : keys) {
                            String value0 = parameters.get(key);
                            String value1 = contentType.parameters.get(key);
                            boolean objectsEqual = value0.equals(value1);
                            if (!objectsEqual)
                            {
                                return false;
                            }
                        }
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
            }
        }
        else
        {
            return false;
        }
    }

    @Override
    public QContentType clone()
    {
        try
        {
            QContentType newObject = (QContentType) super.clone();
            return newObject;
        }
        catch (CloneNotSupportedException e)
        {
            throw new AssertionError(e);
        }
    }
}
