/* Copyright © 2013 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.http.headers;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

/**
 * A parsed accept encoding header field value with a qualification value.
 * <P>
 * The to string value is calculated when first accessed and cached subsequently. The value is checked and set
 * atomically. Concurrent access may cause different String objects to be cached and returned but they will be equal. A
 * happens before relationship is not established by calls to {@link toString()}.
 * 
 * @author Matt Champion
 * @since 0.1.2
 */
public class QEncoding implements Qualified, HeaderFieldElement, Cloneable
{

    private static final int HASH_SEED = 11;
    private static final int HASH_MULT = 17;
    private static final String WILDCARD = "*";

    private final String encoding;
    private final double qualifier;

    private String toString;

    /**
     * Constructor for qualified encodings.
     * @param encoding The encoding
     * @param qualifier The qualifier of the encoding
     * @throws IllegalArgumentException If the encoding is null or if the qualifier is less than
     * zero or greater than one
     */
    public QEncoding(final String encoding, final double qualifier) throws IllegalArgumentException
    {
        if (encoding == null)
        {
            throw new IllegalArgumentException("Encoding must not be null.");
        }
        else if (qualifier < 0.0 || qualifier > 1.0)
        {
            throw new IllegalArgumentException("Qualifier must be between 0.0 and 1.0");
        }
        else
        {
            this.encoding = encoding;
            this.qualifier = qualifier;
        }
    }

    /**
     * Parser for HttpServletRequest to QEncoding
     * 
     * @param request The HTTP request to parse
     * @return The List of encodings in the accept-encoding header
     * @throws UnParsableHeaderException
     * @since 0.1.2
     */
    public static List<QEncoding> getRequestEncodings(HttpServletRequest request)
            throws UnParsableHeaderException
    {
        @SuppressWarnings("unchecked")
        Enumeration<String> headerNames = request.getHeaderNames();
        EncodingParser parser = new EncodingParser();
        List<QEncoding> encodings = new ArrayList<QEncoding>();
        while (headerNames.hasMoreElements())
        {
            String headerName = headerNames.nextElement();
            if (parser.isCorrectHeader(headerName))
            {
                String headerFieldValue = request.getHeader(headerName);
                encodings.addAll(parser.parseAll(headerFieldValue));
            }
        }
        return encodings;
    }

    @Override
    public double getQualifier()
    {
        return qualifier;
    }

    @Override
    public String toString()
    {
        String toString = this.toString;
        if (toString != null)
        {
            return toString;
        }
        else
        {
            toString = encoding + ";q=" + qualifier;
            this.toString = toString;
            return toString;
        }
    }

    /**
     * Method that returns true if the two content types can be matched against one another. Takes
     * into account wildcard matching.
     * 
     * @param encodingObject
     *            Another encoding
     * @return True if they can be matched
     * @since 0.1.2
     */
    public boolean sameEncoding(QEncoding encodingObject)
    {
        if (encodingObject == null)
        {
            return false;
        }
        else
        {
            return this.encoding.equalsIgnoreCase(encodingObject.encoding)
                    || this.encoding.equals(WILDCARD) || encodingObject.encoding.equals(WILDCARD);
        }
    }

    /**
     * Test to see if the encoding is a wildcard.
     * 
     * @return True if the encoding is a wild card.
     * @since 0.1.2
     */
    public boolean isAnyEncoding()
    {
        return WILDCARD.equals(encoding);
    }

    @Override
    public int hashCode()
    {
        int hashCode = HASH_SEED;
        hashCode = (hashCode * HASH_MULT) + encoding.hashCode();
        long temp = Double.doubleToLongBits(qualifier);
        hashCode = (hashCode * HASH_MULT) + (int) (temp ^ (temp >>> 32));
        return hashCode;
    }

    @Override
    public boolean equals(Object object)
    {
        if (object == null)
        {
            return false;
        }
        else if (this == object)
        {
            return true;
        }
        else if (object instanceof QEncoding)
        {
            QEncoding encoding = (QEncoding) object;
            return this.encoding.equals(encoding.encoding)
                    && this.qualifier == encoding.qualifier;
        }
        else
        {
            return false;
        }
    }

    @Override
    public QEncoding clone()
    {
        try
        {
            QEncoding newObject = (QEncoding) super.clone();
            return newObject;
        }
        catch (CloneNotSupportedException e)
        {
            throw new AssertionError(e);
        }
    }
}
