/* Copyright © 2012 Matthew Champion
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 * Neither the name of mattunderscore.com nor the
 names of its contributors may be used to endorse or promote products
 derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.http.headers;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

/**
 * A parsed accept-language header field value with a qualification value.
 * <P>
 * The to string value is calculated when first accessed and cached subsequently. The value is checked and set
 * atomically. Concurrent access may cause different String objects to be cached and returned but they will be equal. A
 * happens before relationship is not established by calls to {@link toString()}.
 * 
 * @author Matt Champion
 * @since 0.0.13
 */
public class QLanguage implements HeaderFieldElement, Qualified, Cloneable
{
    private static final int HASH_SEED = 3;
    private static final int HASH_MULT = 7;
    private static final String WILDCARD = "*";

    private final String language;
    private final String subLanguage;
    private final double qualifier;

    private String toString;

    /**
     * Constructor for qualified languages.
     * @param language The language
     * @param subLanguage The sublanguage
     * @param qualifier The qualifier of the language
     * @throws IllegalArgumentException If the language or sublanguage is null or if the qualifier
     * is less than zero or greater than one
     */
    public QLanguage(String language, String subLanguage, double qualifier) throws IllegalArgumentException
    {
        if (language == null)
        {
            throw new IllegalArgumentException("Language must not be null.");
        }
        else if (subLanguage == null)
        {
            throw new IllegalArgumentException("Sublanguage must not be null.");
        }
        else if (qualifier < 0.0 || qualifier > 1.0)
        {
            throw new IllegalArgumentException("Qualifier must be between 0.0 and 1.0");
        }
        else
        {
            this.language = language;
            this.subLanguage = subLanguage;
            this.qualifier = qualifier;
        }
    }

    /**
     * Parser for HttpServletRequest to QLanguage.
     * 
     * @param request
     *            The HTTP request to parse
     * @return The list of languages in the accept-language header
     * @throws UnParsableHeaderException
     * @since 0.0.13
     */
    public static List<QLanguage> getRequestLanguages(HttpServletRequest request)
            throws UnParsableHeaderException
    {
        @SuppressWarnings("unchecked")
        Enumeration<String> headerNames = request.getHeaderNames();
        LanguageParser parser = new LanguageParser();
        List<QLanguage> types = new ArrayList<QLanguage>();
        while (headerNames.hasMoreElements())
        {
            String headerName = headerNames.nextElement();
            if (parser.isCorrectHeader(headerName))
            {
                String headers = request.getHeader(headerName);
                types.addAll(parser.parseAll(headers));
            }
        }
        return types;
    }

    @Override
    public String toString()
    {
        String toString = this.toString;
        if (toString != null)
        {
            return toString;
        }
        else
        {
            if (WILDCARD.equals(subLanguage))
            {
                toString = language + ";q=" + qualifier;
                this.toString =  toString;
                return toString;
            }
            else
            {
                toString = language + "-" + subLanguage + ";q=" + qualifier;
                this.toString = toString;
                return toString;
            }
        }
    }

    @Override
    public double getQualifier()
    {
        return qualifier;
    }

    /**
     * Test to see if two accept-language header values match. Supports wildcard
     * matches.
     * 
     * @param lang
     * @return True if match
     * @since 0.0.13
     */
    public boolean sameLangauge(QLanguage lang)
    {
        if (this.language.equals(lang.language)
                && this.subLanguage.equals(lang.subLanguage))
        {
            return true;
        }
        else if (this.language.equals(lang.language)
                && (WILDCARD.equals(this.subLanguage) || WILDCARD
                        .equals(lang.subLanguage)))
        {
            return true;
        }
        else
        {
            return WILDCARD.equals(this.language) || WILDCARD.equals(lang.language);
        }
    }

    @Override
    public int hashCode()
    {
        int hashCode = HASH_SEED;
        hashCode = (hashCode * HASH_MULT) + language.hashCode();
        hashCode = (hashCode * HASH_MULT) + subLanguage.hashCode();
        long temp = Double.doubleToLongBits(qualifier);
        hashCode = (hashCode * HASH_MULT) + (int) (temp ^ (temp >>> 32));
        return hashCode;
    }

    @Override
    public boolean equals(Object object)
    {
        if (object == null)
        {
            return false;
        }
        else if (this == object)
        {
            return true;
        }
        else if (object instanceof QLanguage)
        {
            QLanguage qLangunage = (QLanguage) object;
            return (this.language.equals(qLangunage.language)
                    && this.subLanguage.equals(qLangunage.subLanguage) 
                    & this.qualifier == qLangunage.qualifier);
        }
        else
        {
            return false;
        }
    }

    @Override
    public QLanguage clone()
    {
        try
        {
            QLanguage newObject = (QLanguage) super.clone();
            return newObject;
        }
        catch (CloneNotSupportedException e)
        {
            throw new AssertionError(e);
        }
    }
}
